//Copyright (C) 2011 Pierre Moulon, 2016 Sinia Petri

#ifndef SP_IMAGE_H
#define SP_IMAGE_H

#include <iostream>
#include <cstring>
using namespace std;

//-- Container for a 2D image
//-- This class ensure that the image have a width and a height
//-- and a 2D array of T data.
//-
//-- Data is saved in row major format
//-- Pixel access is done with operator(y,x)
//  [2/3/2011 pierre MOULON]
//---------------------------
enum speEdgeOptions {eopSkip, eopWrap, eopBorder};
enum speOrientation {orAsIs, orInvert};

template <typename T> class spImage
  {
  public:
	typedef T Tpixel;
  speEdgeOptions PixelOverFlow;
  //------------------------------
  // constructor - allocate image width height
  //------------------------------
  spImage (size_t width, size_t height, bool contiguous = false)
	{
	_height = height;
	_width = width;
	_allocateMemory(contiguous, sizeof(T));
	_shared = false;
	PixelOverFlow = eopBorder;
	}
  //------------------------------
  // constructor - copy image
  //------------------------------
  spImage(const spImage<T> *I)
	{
	T **inData = I->_data;
	_height    = I->_height;
	_width     = I->_width;
	_contiguous_memory = I->_contiguous_memory;
	_allocateMemory((bool)_contiguous_memory, sizeof(T));
	// copy  data
	for (size_t i = 0; i < _height; i++)
		 memcpy(_data[i], I->_data[i], _width * sizeof(T));
	PixelOverFlow = I->PixelOverFlow;
	_shared = false;
	}
  //------------------------------
  // constructor - share image
  //------------------------------
  spImage(void **scanLineData, size_t width, size_t height)
	{
	_width = width;
	_height = height;
	_shared = true;
	_data = new T*[_height];
	for (size_t i = 0; i < _height; i++)
		_data[i] = static_cast <T*>(scanLineData[i]);
	PixelOverFlow = eopBorder;
	}
  //------------------------------
  // constructor - share image
  //------------------------------
  spImage(Byte **scanLineData, size_t width, size_t height)
	{
	_width = width;
	_height = height;
	_contiguous_memory = 0;
	_shared = true;
	_data = new T*[_height];
	for (size_t i = 0; i < _height; i++)
		_data[i] = reinterpret_cast <T*>(scanLineData[i]);
	PixelOverFlow = eopBorder;
	}
  //------------------------------
  // constructor - raw image
  //------------------------------
  spImage(void *buffer, size_t width, size_t height, unsigned int scanlineAlignment, bool shared = true, speOrientation orientation = orAsIs)
	{
	_width = width;
	_height = height;
	_shared = shared;
	int stride = _width * sizeof(T);
	int align = stride % scanlineAlignment;         // alignemnt
	if (align > 0) align =  scanlineAlignment - align;
	stride = stride + align; 					    // stride
	void *currPtr = buffer;
	if (!_shared)
	   _allocateMemory(true, sizeof(T));	// always contiguous
	else
	   _data = new T*[_height];
	if (orientation == orAsIs)
	   {
	   for (size_t i = 0; i < _height; i++)
		   {
		   _data[i] = static_cast <T*>(currPtr);
		   currPtr = static_cast<void*>((Byte*)currPtr + stride);
		   }
	   }
	else
	   {
	   for (size_t i = _height-1; i >= 0; i++)
		   {
		   _data[i] = static_cast <T*>(currPtr);
		   currPtr = static_cast<void*>((Byte*)currPtr - stride);
		   }
       }
	PixelOverFlow = eopBorder;
	}
  //------------------------------
  // destructor
  //------------------------------
  virtual ~spImage()
	{
	_freeMemory();
	}
  //------------------------------
  // accessors/getters methods
  //------------------------------
  // Retrieve the width of the image
  size_t Width()  const { return _width; }
  // Retrieve the height of the image
  size_t Height() const { return _height; }
  // Return the depth in byte of the pixel (unsigned char will return 1)
  size_t Depth() const  { return sizeof(Tpixel);}
  // Return the depth in byte of the pixel (unsigned char will return 1)
  bool IsContiguous() const  { return _contiguous;}
  //------------------------------
  // image memory access
  //------------------------------
  T ** Data() const {return _data;}
  // random pixel access
  inline const T & operator()(size_t y, size_t x) const {return _data[y][x];}
  // random pixel access
  inline T & operator()(size_t y, size_t x) {return _data[y][x];}
  // random pixel access
  inline const T & Pixel(size_t y, size_t x) const {return _data[y][x];}
  // random pixel access
  inline T & Pixel(size_t y, size_t x) {return _data[y][x];}
  // are imagea equal
  inline bool operator==(const spImage<T> *img) const
	{
	if (_width == img->_width && _height == img->_height)
	   {
	   for (size_t i = 0; i < _height; i++)
		   {
		   if (0 != memcmp(_data[i], img->_data[i], _width)
			  return false;
		   }
	   return true;
	   }
	return false;
	}
  //------------------------------
  // accessors/getters methods
  //------------------------------

  // Tell if a point is inside the image.
  bool Contains(size_t y, size_t x) const {
	return x < _width && y < _height;
  }
  // Tell if a point is inside the image.
  bool Contains(double y, double x) const {
	return (x >=0 && x <_width && y >=0 && y <_height);
  }
  // Utils
  //------------------------------
  // Fill with color
  //------------------------------
  void Fill(const T &val)
  {
	for (int i = 0; i < _height; i++)
		{
		for (int j = 0; j < _width; j++)
			_data[i][j] = val;
		}
	//memset(_data, val, sizeof(T) * _width * _height);
  }
  //------------------------------
  // copy image
  //------------------------------
  void Copy(const spImage<T> *I)
	{
	T **inData = I->_data;
	size_t height = min(_height, I->_height);
	size_t width  = min(_width,  I->_width);
	// copy  data
	for (size_t i = 0; i < height; i++)
		 memcpy(_data[i], I->_data[i], width * sizeof(T));
	}
  //------------------------------
  // Set safe index
  //------------------------------
  bool SetSafeIndex (int *index, int *jndex)
	{
	if (*index > 0 && *index < _height && *jndex > 0 && *jndex < _width)
	   return true;
	if (PixelOverFlow == eopWrap)
	   {
	   if (*index < 0)
		  *index = _height + *index;
	   else if ( *index >= _height)
		  *index =  *index - _height;
	   if (*jndex < 0)
		  *jndex = _width + *jndex;
	   else if ( *jndex >= _width)
		  *jndex =  *jndex - _width;
	   }
	else if (PixelOverFlow == eopBorder)
	   {
	   if (*index < 0)
		  *index = 0;
	   else if ( *index >= _height)
		  *index =  _height - 1;
	   if (*jndex < 0)
		  *jndex = 0;
	   else if ( *jndex >= _width)
		  *jndex =  _width - 1;
	   }
	else   //eopSkip
	   return false;
	return true;
	}
protected :
  // Internal data :
  T ** _data;
  T *_contiguous_memory;
  size_t _width;
  size_t _height;
  bool _shared;
  // Internal methods
  //------------------------------
  void _allocateMemory(bool contiguous, size_t pix_size)
	{
	const size_t n = _width*_height;
	//const size_t pix_size = sizeof(T);
	_data = new T*[_height];
	if (contiguous) // contiguous memory
	   {
	   _contiguous_memory = static_cast<T*>(GetMemory(n * pix_size));
	   for (size_t i = 0; i < _height; i++)
		   _data[i] = (_contiguous_memory + (i * _width));
	   }
	else 	// discontiguous memory
	   {
	   _contiguous_memory = 0;
	   for (size_t i = 0; i < _height; i++)
		   _data[i] = static_cast<T*>(GetMemory(_width * pix_size));
	   }
	}
  //------------------------------
  void _freeMemory(void)
	{
	if (!_data)
	   return;
	if (_shared)
	   {
	   for (size_t i = 0; i < _height; i++)
		   _data[i] = 0;
	   }
	else if (_contiguous_memory)
	   {
	   for (size_t i = 0; i < _height; i++)
		   _data[i] = 0;
	   FreeMemory(_contiguous_memory);
	   }
	else
	   {
	   for (size_t i = 0; i < _height; i++)
		   {
		   FreeMemory(_data[i]);
		   _data[i] = 0;
		   }
	   }
	delete [] _data;
	_data = 0;
	_contiguous_memory = 0;
	_width = _height = 0;
	}
};
#endif // SP_IMAGE_H

